#!/usr/local/bin/python3.10

import ipwhois
import os
import pprint
import shelve
import signal
import sys

# WHOIS time out in seconds
WHOIS_TIMEOUT = 15

################################################################################
def _handle_timeouts(signal_number, current_stack_frame):
    raise Exception("Timeout!")

################################################################################
def _whois_lookup(ip):
    """Perform an WHOIS RDAP lookup"""
    signal.alarm(WHOIS_TIMEOUT)
    try:
        # Try to resolve from WHOIS:
        object = ipwhois.IPWhois(ip)
        data = object.lookup_rdap(depth=1)
    except Exception:
        # No WHOIS answer in WHOIS_TIMEOUT seconds
        data = {}
    signal.alarm(0)

    data["query"] = ip

    return data

################################################################################
def _store_whois_results(ip, data):
    """Store or update a bulky WHOIS RDAP data structure in a shelf DB"""
    records = 0
    with shelve.open("whois_cache") as db:
        if ip in db:
            print("Updating {:s}".format(ip))
        else:
            print("Adding {:s}".format(ip))
            records = 1
        db[ip] = data

    return records

################################################################################
def _list_size(data):
    """"""
    size = 0
    for item in data:
        if isinstance(item, str):
            size += len(item)
        elif isinstance(item, dict):
            size += _dict_size(item)
        elif isinstance(item, list):
            size += _list_size(value)
        elif item == None:
            pass
        else:
            print(type(item))

    return size

################################################################################
def _dict_size(data):
    """"""
    size = 0
    for key, value in data.items():
        size += len(key)
        if isinstance(value, str):
            size += len(value)
        elif isinstance(value, dict):
            size += _dict_size(value)
        elif isinstance(value, list):
            size += _list_size(value)
        elif value == None:
            pass
        else:
            print(type(value))

    return size

################################################################################
def _print_db_size(records, data):
    """"""
    #sys.getsizeof(data)
    print(
        "Database has {:d} records for {:d} bytes. Last record was {:d} bytes long".format(
            records,
            os.path.getsize("whois_cache.db"),
            _dict_size(data),
        )
    )

################################################################################
def _verify_whois_cache():
    """Just go through all the DB keys to see if it blows up"""
    with shelve.open("whois_cache") as db:
        for key in db.keys():
            pass

################################################################################
signal.signal(signal.SIGALRM, _handle_timeouts)

# processing an IP list 1 by 1 to see at which point the Shelve database becomes unreadable
records = 0
with open("ip", "r") as file:
    ips = file.readlines()
    for ip in ips:
        ip = ip.strip()

        data = _whois_lookup(ip)

        # Make a text copy of the RDAP WHOIS data
        with open("whois_cache.txt", "a") as file:
            pprint.pprint(data, file)

        records += _store_whois_results(ip, data)
        _print_db_size(records, data)
        _verify_whois_cache()

